// hdstandby copyright 2003-2005 by Charl P. Botha http://cpbotha.net/
// portions of the code copyright Microsoft

// this will set the idle timeout on your first drive to 0 (i.e. no timeout) - change the line 
// "strncpy(driveSpec, "\\\\.\\PhysicalDrive0", 1023);" to apply this to another drive
// this solves the problem where some drives experience a hardware power saving timeout on 
// windows in spite of the fact that no power saving timeout is set; it is equivalent to hdparm -S 0 on linux
// this problem manifests itself in mini "freezes" whilst the hard drive spins up from a timeout; not even
// mouse motion registers during these moments.

// this was tested on XP SP1; I got most of my info from the hdparm source code and from some Microsoft
// examples; very complete information on how to use IOCTL_IDE_PASS_THROUGH is also available at:
// ftp://ftp.heise.de/pub/ct/listings/0207-218.zip

// WARNING:
// USE THIS CODE AT YOUR OWN RISK;  IT WORKS FOR ME ON ONE SPECIFIC LAPTOP, I HAVEN'T
// TESTED IT ANYWHERE ELSE. 

#include <stdio.h>
#include <conio.h>
#include <windows.h>

#pragma pack(1)   // Required to ensure correct SMART IOCTL structure setup
#include "ataPassThroughIOCTL.h"
#include "smart.h"
//#include "smartapp.h"

//---------------------------------------------------------------------
//---------------------------------------------------------------------
VOID ChangeByteOrder(PCHAR szString, USHORT uscStrSize)
{

  USHORT        i;
  CHAR  temp;

  for (i = 0; i < uscStrSize; i+=2)
    {
      temp = szString[i];
      szString[i] = szString[i+1];
      szString[i+1] = temp;
    }
}

//---------------------------------------------------------------------
// Display contents of IDE hardware registers reported by SMART
//---------------------------------------------------------------------
VOID PrintIDERegs(PSENDCMDINPARAMS pscip)
{
  printf("\tIDE TASK FILE REGISTERS:\n");

  printf("\t\tbFeaturesReg     = 0x%X\n", pscip->irDriveRegs.bFeaturesReg);
  printf("\t\tbSectorCountReg  = 0x%X\n", pscip->irDriveRegs.bSectorCountReg);
  printf("\t\tbSectorNumberReg = 0x%X\n", pscip->irDriveRegs.bSectorNumberReg);
  printf("\t\tbCylLowReg       = 0x%X\n", pscip->irDriveRegs.bCylLowReg);
  printf("\t\tbCylHighReg      = 0x%X\n", pscip->irDriveRegs.bCylHighReg);
  printf("\t\tbDriveHeadReg    = 0x%X\n", pscip->irDriveRegs.bDriveHeadReg); 
  printf("\t\tStatus           = 0x%X\n", pscip->irDriveRegs.bCommandReg);
}

void interpretStandby (BYTE standby)
{

  switch(standby) {
  case 0:               printf("off");
    break;
  case 252:     printf("21 minutes");
    break;
  case 253:     printf("vendor-specific");
    break;
  case 254:     printf("?reserved");
    break;
  case 255:     printf("21 minutes + 15 seconds");
    break;
  default:
    if (standby <= 240) {
      unsigned int secs = standby * 5;
      unsigned int mins = secs / 60;
      secs %= 60;
      if (mins)   printf("%u minutes", mins);
      if (mins && secs) printf(" + ");
      if (secs)   printf("%u seconds", secs);
    } else if (standby <= 251) {
      unsigned int mins = (standby - 240) * 30;
      unsigned int hrs  = mins / 60;
      mins %= 60;
      if (hrs)    printf("%u hours", hrs);
      if (hrs && mins)  printf(" + ");
      if (mins)   printf("%u minutes", mins);
    } else
      printf("illegal value");
    break;
  }

}

/****************************************************************************
 *
 * DoEnableSMART
 *
 * FUNCTION: Send a SMART_ENABLE_SMART_OPERATIONS command to the drive
 * bDriveNum = 0-3
 *
 ****************************************************************************/
BOOL doSetStandby2(BYTE standbyValue, HANDLE hDRIVE, 
                  PSENDCMDINPARAMS pSCIP, PSENDCMDOUTPARAMS pSCOP, BYTE bDriveNum, PDWORD lpcbBytesReturned)
{
#define WIN_SETIDLE1 0xE3

  pSCIP->cBufferSize = 0;

  pSCIP->irDriveRegs.bFeaturesReg = 0x0; 
  // from section 8.17.4 of the ATA spec, it looks like the standbyValue
  // should go here, see http://www.t13.org/project/d1410r3a.pdf
  pSCIP->irDriveRegs.bSectorCountReg = standbyValue; 
  pSCIP->irDriveRegs.bSectorNumberReg = 0x0; 
  pSCIP->irDriveRegs.bCylLowReg = 0x0;
  pSCIP->irDriveRegs.bCylHighReg = 0x0;

  //
  // Compute the drive number.
  // According to the ATA spec, this is how the device is built up...
  //
  pSCIP->irDriveRegs.bDriveHeadReg = 0xA0 | ((bDriveNum & 1) << 4); 
  // bCommandReg == actual drive command
  pSCIP->irDriveRegs.bCommandReg = WIN_SETIDLE1; 
  pSCIP->bDriveNumber = bDriveNum;

  // the SPEC says something about Prerequisites: DRDY should be 1 ???
  //pSCOP->DriverStatus.status = 64;

  printf("sizeof(pSCOP) == %d\n", sizeof(SENDCMDOUTPARAMS));
  return (DeviceIoControl(hDRIVE, DFP_SEND_DRIVE_COMMAND,
                          (LPVOID)pSCIP, sizeof(SENDCMDINPARAMS) - 1,
                          (LPVOID)pSCOP, sizeof(SENDCMDOUTPARAMS) - 1,
                          lpcbBytesReturned, NULL) );

}



BOOL doSetStandby(BYTE standbyValue, HANDLE hDRIVE, int driveNum, PDWORD lpcbBytesReturned)
                  
{
  
#define WIN_SETIDLE1 0xE3
// I have NO idea about this value, I'm hoping this is okay...
#define ATA_FLAGS_DRDY_REQUIRED 0x01

  ATA_PASS_THROUGH apte;

  // zero everything to begin with
  memset(&apte, 0, sizeof(apte));

  // from section 8.17.4 of the ATA spec, it looks like the standbyValue
  // should go here, see http://www.t13.org/project/d1410r3a.pdf
  // in SectorCount
  apte.IdeReg.bSectorCountReg = standbyValue;
  
  // According to the ATA spec, this is how the device is built up...
  apte.IdeReg.bDriveHeadReg = 0xA0 | ((driveNum & 1) << 4);

  // command (IDLE1, 0xE3) goes here
  apte.IdeReg.bCommandReg = WIN_SETIDLE1;

  //apte.Length = sizeof(apte);
  // according to the spec, this one requires DRDY to be 1
  //apte.AtaFlags = ATA_FLAGS_DRDY_REQUIRED;


  return (DeviceIoControl(hDRIVE, IOCTL_IDE_PASS_THROUGH,
                          (LPVOID)&apte, sizeof(apte),
                          (LPVOID)&apte, sizeof(apte),
                          lpcbBytesReturned, NULL) );

}


/****************************************************************************
 *
 * DoIDENTIFY
 *
 * FUNCTION: Send an IDENTIFY command to the drive
 * bDriveNum = 0-3
 * bIDCmd = IDE_ID_FUNCTION or IDE_ATAPI_ID
 *
 ****************************************************************************/
BOOL DoIDENTIFY(HANDLE hDRIVE, PSENDCMDINPARAMS pSCIP,
                PSENDCMDOUTPARAMS pSCOP, BYTE bIDCmd, BYTE bDriveNum, 
                PDWORD lpcbBytesReturned)
{
  //
  // Set up data structures for IDENTIFY command.
  //

  pSCIP->cBufferSize = IDENTIFY_BUFFER_SIZE;

  pSCIP->irDriveRegs.bFeaturesReg = 0;
  pSCIP->irDriveRegs.bSectorCountReg = 1;
  pSCIP->irDriveRegs.bSectorNumberReg = 1;
  pSCIP->irDriveRegs.bCylLowReg = 0;
  pSCIP->irDriveRegs.bCylHighReg = 0;

  //
  // Compute the drive number.
  //
  pSCIP->irDriveRegs.bDriveHeadReg = 0xA0 | ((bDriveNum & 1) << 4); 

  //
  // The command can either be IDE identify or ATAPI identify.
  //
  pSCIP->irDriveRegs.bCommandReg = bIDCmd;
  pSCIP->bDriveNumber = bDriveNum;
  pSCIP->cBufferSize = IDENTIFY_BUFFER_SIZE;

  return ( DeviceIoControl(hDRIVE, DFP_RECEIVE_DRIVE_DATA,
                           (LPVOID)pSCIP, sizeof(SENDCMDINPARAMS) - 1,
                           (LPVOID)pSCOP, 
                           sizeof(SENDCMDOUTPARAMS) + IDENTIFY_BUFFER_SIZE - 1,
                           lpcbBytesReturned, NULL) );
}

/****************************************************************************
 *
 * DisplayIdInfo
 *
 * Display the contents of the ID buffer
 *
 ****************************************************************************/
VOID DisplayIdInfo(PIDSECTOR pids, PSENDCMDINPARAMS pSCIP, 
                   BYTE bIDCmd, BYTE bDfpDriveMap, BYTE bDriveNum)
{
  BYTE  szOutBuffer[41];

  if (bIDCmd == IDE_ID_FUNCTION)
    {
      printf("Drive %d is an IDE Hard drive", bDriveNum); 
      printf("%s\n", ( bDfpDriveMap >> bDriveNum & 1) ? \
             " that supports SMART" : "");
      printf("\t#Cylinders: %d, #Heads: %d, #Sectors per Track: %d\n",
             pids->wNumCyls, 
             pids->wNumHeads, 
             pids->wSectorsPerTrack);

      //PrintIDERegs(pSCIP);

    }
  else
    {
      printf("Drive %d is an ATAPI device.\n", bDriveNum); 
    }

  //
  // Change the WORD array to a BYTE
  // array
  //
  ChangeByteOrder(pids->sModelNumber, 
                  sizeof pids->sModelNumber);

  memset(szOutBuffer,0, sizeof(szOutBuffer));
  strncpy(szOutBuffer, pids->sModelNumber, sizeof(pids->sModelNumber));

  printf("\tModel number: %s\n", szOutBuffer);


  //
  // Change the WORD array to a BYTE
  // array
  //
  ChangeByteOrder(pids->sFirmwareRev, 
                  sizeof pids->sFirmwareRev);

  memset(szOutBuffer,0, sizeof(szOutBuffer));
  strncpy(szOutBuffer, pids->sFirmwareRev, sizeof(pids->sFirmwareRev));

  printf("\tFirmware rev: %s\n", szOutBuffer);


  //
  // Change the WORD array to a BYTE
  // array
  //
  ChangeByteOrder(pids->sSerialNumber, 
                  sizeof pids->sSerialNumber);

  memset(szOutBuffer,0, sizeof(szOutBuffer));
  strncpy(szOutBuffer, pids->sSerialNumber, sizeof(pids->sSerialNumber));

  printf("\tSerial number: %s\n",szOutBuffer );
       
}




// drivespec == e.g. \\\\.\\PhysicalDrive0
HANDLE openDrive(const char *driveSpec)
{
  HANDLE        hSMARTIOCTL = 0;
        
  // Windows NT, Windows 2000
  if ((hSMARTIOCTL = CreateFile(driveSpec,GENERIC_READ | GENERIC_WRITE,FILE_SHARE_READ|FILE_SHARE_WRITE,NULL,
                                OPEN_EXISTING,0,NULL)) == INVALID_HANDLE_VALUE)
    {
      printf("Unable to open physical drive, error code: 0x%lX\n", GetLastError());
    }
  else
    {
      printf("Physical drive opened successfully\n");
    }

  return hSMARTIOCTL;

}


//
// Define global buffers.
//

BYTE    IdOutCmd[sizeof(SENDCMDOUTPARAMS) + IDENTIFY_BUFFER_SIZE - 1];

int main(int argc, char *argv[])
{
  HANDLE                hDRIVE = 0;
  DWORD                 cbBytesReturned;
  SENDCMDINPARAMS       scip;
  SENDCMDOUTPARAMS      OutCmd;
  BYTE                          bDfpDriveMap = 0;
  BYTE                          bSuccess = 1;
  BYTE                          bIDCmd;         // IDE or ATAPI IDENTIFY cmd

  BYTE                standbyValue;
  BYTE                driveNum;
  unsigned char       driveSpec[1024];

  printf("hdstandby copyright (c) 2003 by Charl P. Botha\n");
  printf("          portions copyright Microsoft\n\n");
  printf("USE HDSTANDBY AT YOUR OWN RISK!\n\n");


  standbyValue = 255;
  driveNum = 0;
  strncpy(driveSpec, "\\\\.\\PhysicalDrive0", 1023);
  
  printf("Attempting to set ATA standby of physical drive %d to ", driveNum);
  interpretStandby(standbyValue);
  printf("\n\n");

  //
  // Try to get a handle to SMART IOCTL, report failure and exit if
  // can't.
  // \\\\.\\PhysicalDrive0
  if ((hDRIVE = openDrive(driveSpec)) != INVALID_HANDLE_VALUE)
    {

      memset(&scip, 0, sizeof(scip));
      memset(&OutCmd, 0, sizeof(OutCmd));
  
      if ( DoIDENTIFY(hDRIVE, 
                      &scip, 
                      (PSENDCMDOUTPARAMS)&IdOutCmd, 
                      IDE_ID_FUNCTION,
                      driveNum, 
                      &cbBytesReturned))
        {
          DisplayIdInfo((PIDSECTOR) ((PSENDCMDOUTPARAMS)IdOutCmd)->bBuffer,
                        &scip,
                        IDE_ID_FUNCTION,
                        bDfpDriveMap,
                        driveNum);
        }

      else
        {
          printf("Identify Command Failed on Drive: %d\n", driveNum);
          printf(" DriverStatus: error=0x%X, status=0x%X\n\n", 
                 ((PSENDCMDOUTPARAMS)IdOutCmd)->DriverStatus.error, 
                 ((PSENDCMDOUTPARAMS)IdOutCmd)->DriverStatus.status);
        }
      printf("\tcbBytesReturned: %d\n\n", cbBytesReturned);
    }


  memset(&scip, 0, sizeof(scip));
  memset(&OutCmd, 0, sizeof(OutCmd));
  cbBytesReturned = 0;
  //if (doSetStandby(standbyValue, hDRIVE, &scip, &OutCmd, driveNum, &cbBytesReturned)) 
  if (doSetStandby(standbyValue, hDRIVE, driveNum, &cbBytesReturned))
    {
      printf("Successfully set drive standby.\n");
    }
  else
    {
      printf("Error setting drive standby!\n");
      /*printf(" DriverStatus: error=0x%X, status=0x%X\n\n", 
               OutCmd.DriverStatus.error, 
               OutCmd.DriverStatus.status);*/
      printf("BytesReturned == %d, LastError == %d\n", cbBytesReturned, GetLastError());
      // 87 	The parameter is incorrect. 	ERROR_INVALID_PARAMETER
    }

  CloseHandle(hDRIVE);


  return(0);
}
